"""Module containg data types imported throughout."""

import enum
from typing import List, Literal, Union
from pydantic import BaseModel, Field, field_serializer


class ReleaseType(enum.Enum):
    """Logical type of semantic versioning release."""

    NO_RELEASE = "no release"
    PATCH = "patch"
    MINOR = "minor"
    MAJOR = "major"


class GitConfig(BaseModel):
    """Git configuration options to set before committing."""

    name: str
    email: str


class Prefix(BaseModel):
    """Commit prefix that can trigger releases."""

    label: str
    type: ReleaseType

    @field_serializer("type")
    def serialize_type(self, type: ReleaseType, _info):
        """Convert ReleaseType to a string when serializing."""
        return type.value


class ReplaceFileType(str, enum.Enum):
    """File types supported for `replace_files` in config."""

    JSON = "json"
    SED = "sed"
    TOML = "toml"
    YAML = "yaml"


class ReplaceFile(BaseModel):
    """File to be replaced on new release."""

    name: str
    type: ReplaceFileType

    @field_serializer("type")
    def get_enum_value(self, v, info) -> str:
        """Turn the type field into a string when serializing."""
        return str(v.value)


class JsonReplaceFile(ReplaceFile):
    """Represents JSON file to replaced on new release."""

    type: Literal[ReplaceFileType.JSON] = ReplaceFileType.JSON
    key: str


class SedReplaceFile(ReplaceFile):
    """Represents a file to replaced using `sed` on new release."""

    type: Literal[ReplaceFileType.SED] = ReplaceFileType.SED
    script: str


class TomlReplaceFile(ReplaceFile):
    """Represents TOML file to replaced on new release."""

    type: Literal[ReplaceFileType.TOML] = ReplaceFileType.TOML
    key: str


class YamlReplaceFile(ReplaceFile):
    """Represents YAML file to replaced on new release."""

    type: Literal[ReplaceFileType.YAML] = ReplaceFileType.YAML
    key: str


ReplaceFileUnion = Union[
    JsonReplaceFile, SedReplaceFile, TomlReplaceFile, YamlReplaceFile
]


class SemverConfig(BaseModel):
    """Represents config options for entire application."""

    changelog_path: str
    changelog_writer: str
    format: str
    git: GitConfig
    prefixes: List[Prefix]
    replace_files: List[ReplaceFileUnion] = Field(discriminator="type")



"""Module related to config files."""

import os
from pathlib import Path

import yaml

from pagekey_semver.models import GitConfig, Prefix, SemverConfig
from pagekey_semver.util.env_to_dict import convert_env_to_dict
from pagekey_semver.util.update_dict import merge_dicts


DEFAULT_CONFIG = SemverConfig(
    changelog_path="CHANGELOG.md",
    changelog_writer="pagekey_semver.changelog:DefaultChangelogWriter",
    format="v%M.%m.%p",
    git=GitConfig(
        name="PageKey Semver",
        email="semver@pagekey.io",
    ),
    prefixes=[
        Prefix(label="major", type="major"),
        Prefix(label="minor", type="minor"),
        Prefix(label="feat", type="minor"),
        Prefix(label="fix", type="patch"),
    ],
    replace_files=[],
)
DEFAULT_CONFIG_DICT = DEFAULT_CONFIG.model_dump()


def load_config(config_path: Path) -> SemverConfig:
    """Load config from default, file, environment (in that order).
    
    Args:
        config_path: Path to config to parse.
    
    Returns:
        SemverConfig representing merged config from defaults, config file, and environment settings.
    """
    # Get the file config.
    if config_path.is_file():
        with open(config_path, "r") as file_handle:
            custom_config = yaml.safe_load(file_handle.read())
        config_without_env = merge_dicts(DEFAULT_CONFIG_DICT, custom_config)
    else:
        config_without_env = DEFAULT_CONFIG_DICT

    # Get the config defined by environment variables.
    env_config = convert_env_to_dict(os.environ)

    # Merge the file config and the environment config.
    # Environment takes precedence.
    final_config = merge_dicts(config_without_env, env_config)
    return SemverConfig(**final_config)



"""Module for interacting with Git."""

import os
import subprocess
from typing import List, Optional
from pagekey_semver.models import GitConfig, SemverConfig
from pagekey_semver.release import Commit, Tag


class GitManager:
    """Class to handle all communications with Git executable."""

    def __init__(self, config: SemverConfig):
        """Initialize Git manager."""
        self._config = config

    def get_existing_git_info(self) -> Optional[GitConfig]:
        """Determine exisitng Git config information.
        
        Returns:
            GitConfig containing Git user.name and user.email.
        
        Raises:
            CalledProcessError if there is an issue calling Git to check these values.
        """
        result = subprocess.run(
            ["git", "config", "user.email"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        email = result.stdout.strip()
        result = subprocess.run(
            ["git", "config", "user.name"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        name = result.stdout.strip()
        return GitConfig(name=name, email=email)

    def get_git_tags(self) -> List[str]:
        """Get a list of all Git tags for the current repo.
        
        Returns:
            List of each tag name as a string.
        """
        try:
            result = subprocess.run(
                ["git", "tag"],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            # Split the output into lines (each tag is on a new line)
            tags = result.stdout.strip().split("\n")
            return tags
        except subprocess.CalledProcessError as e:
            print(f"Error getting git tags: {e.stderr}", flush=True)
            return []

    def get_commit_messages_since(self, commit_hash: Optional[str]) -> List[Commit]:
        """Return a  list of commit messages since a commit ref.
        
        Args:
            commit_hash: Reference to a commit to pull messages since, or None.
        
        Returns:
            Commit messages since `commit_hash` if provided.
            All commit messages if `commit_hash` is None.
        """
        try:
            if commit_hash is None:
                args = ["git", "log", "--pretty=format:%H %s"]
            else:
                args = ["git", "log", f"{commit_hash}..HEAD", "--pretty=format:%H %s"]
            result = subprocess.run(
                args,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            # Split the output into lines (each line is a commit message)
            commit_lines = result.stdout.strip().split("\n")
            commits = []
            for line in commit_lines:
                hash = line.split()[0]
                message = line.replace(hash + " ", "")
                commits.append(Commit(hash, message))
            return commits
        except subprocess.CalledProcessError as e:
            print(f"Error getting commit messages: {e.stderr}", flush=True)
            return []

    def apply_tag(self, existing_tags: List[Tag], new_tag: Tag) -> None:
        """Commit, tag, and push.

        Changes Git username, email based on config
        and restores them to what they were previously when done.
        
        Args:
            existing_tags: List of pre-existing Git tags in repo.
            new_tag: Tag to be added, if it does not already exist.
        """
        if new_tag not in existing_tags:
            print(f"Tagging/pushing new tag: {new_tag}", flush=True)
            original_git_config = self.get_existing_git_info()
            commands = [
                f'git config user.email "{self._config.git.email}"',
                f'git config user.name "{self._config.git.name}"',
                f"git add --all",
                f"git commit -m '{new_tag.name}'",
                f"git tag {new_tag.name}",
                f"git push origin {new_tag.name}",
                f"git push origin HEAD",
                f'git config user.email "{original_git_config.email}"',
                f'git config user.name "{original_git_config.name}"',
            ]
            for command in commands:
                print("Running:", command, flush=True)
                exit_code = os.system(command)
                if exit_code != 0:
                    raise ValueError(f"Command failed: {command}")
        else:
            print(f"Tag {new_tag} already exists - skipping tag/push.", flush=True)



"""Module for CLI."""

import argparse
import sys
from pathlib import Path

from pagekey_semver.changelog import ChangelogWriter
from pagekey_semver.file_replacer import FileReplacer
from pagekey_semver.git import GitManager
from pagekey_semver.release import SemverRelease, Tag
from pagekey_semver.config import load_config


def cli_entrypoint(args=sys.argv[1:]):
    """Runs pagekey-semver when called from the command-line.
    
    Args:
        args: List of command-line args passed by user, or fake args for testing purposes.
    """
    parser = argparse.ArgumentParser(description="PageKey Semver")
    subparsers = parser.add_subparsers(title="Commands", dest="command")
    subparsers.add_parser("plan", help="Compute which version would be created. (Dry-run)")
    subparsers.add_parser("apply", help="Compute version, then commit, tag, and push.")
    parsed_args = parser.parse_args(args)

    if parsed_args.command == "plan":
        dry_run = True
    elif parsed_args.command == "apply":
        dry_run = False
    else:
        parser.print_usage()
        return

    config = load_config(Path(".semver"))

    # Init classes.
    manager = GitManager(config)
    release = SemverRelease(config)
    writer = ChangelogWriter.from_config(config)

    # Compute tags, commits
    tags = manager.get_git_tags()
    max_tag: Tag = release.get_biggest_tag(tags)
    hash = None
    if max_tag is not None:
        hash = max_tag.name
    commits = manager.get_commit_messages_since(hash)
    release_type = release.compute_release_type(commits)
    next_version = release.compute_next_version(release_type, tags)

    # Write to changelog.
    writer.update_changelog(next_version, commits)
    print("Next version:", next_version, flush=True)

    # Apply tag if appropriate.
    if not dry_run:
        # Replace files
        file_replacer = FileReplacer(config, next_version)
        file_replacer.replace_all()
        # Apply tag, commit, push
        manager.apply_tag(tags, next_version)
    else:
        print("Dry run mode - not applying version.", flush=True)


if __name__ == "__main__":
    cli_entrypoint()



"""Module to convert env vars to a dict."""

from typing import Dict


VARIABLE_PREFIX = "SEMVER_"


def convert_env_to_dict(variables: Dict[str, str]) -> Dict:
    """Convert variables dict to config dict.

    This function takes a list of flattened config items and
    re-hydrates them into a valid config dict.

    To learn more about the expected inputs and outputs, check out the tests for this function or the docs.
    
    Args:
        variables: Dict with keys/values that represent environment variables.

    Returns:
        Dictionary that can be parsed into a valid SemverConfig object.
    """
    # Filter irrelevant vars and get rid of prefix.
    variables = {
        key.replace(VARIABLE_PREFIX, ""): variables[key]
        for key in variables
        if key.startswith(VARIABLE_PREFIX)
    }

    result = {}

    for variable, value in variables.items():
        parts = variable.split("__")
        # Create a nested dictionary structure
        d = result
        for part in parts[:-1]:
            if part not in d:
                d[part] = {}
            d = d[part]

        # Set the final key's value
        d[parts[-1]] = value

    # Manually re-arrange config items that are lists
    if "prefixes" in result:
        new_prefixes = []
        for key, value in result["prefixes"].items():
            new_prefixes.append(
                {
                    "label": key,
                    "type": value,
                }
            )
        result["prefixes"] = new_prefixes
    if "replace_files" in result:
        # Simply discard the keys to convert to list.
        new_replace_files = list(result["replace_files"].values())
        result["replace_files"] = new_replace_files

    return result



"""Module for computing release logic. related to computing release."""

from dataclasses import dataclass
import re
from typing import List, Optional

from pagekey_semver.models import ReleaseType, SemverConfig


@dataclass
class Commit:
    """Represents a Git commit."""

    hash: str
    message: str


@dataclass
class Tag:
    """Represents a Semver logical tag."""

    # The formatted tag name as it appears in Git.
    name: str
    # The major part of the version.
    major: int
    # The minor part of the version.
    minor: int
    # The patch part of the version.
    patch: int


# This variable assigns a numeric priority to each ReleaseType,
# which allows the higher numbers to take precedence over the
# lower ones.
RELEASE_TYPE_PRIORITIES = {
    ReleaseType.NO_RELEASE: 0,
    ReleaseType.PATCH: 1,
    ReleaseType.MINOR: 2,
    ReleaseType.MAJOR: 3,
}


def release_greater(a: ReleaseType, b: ReleaseType) -> bool:
    """Determine which of the two release types is greater in priority.
    
    Args:
        a: The first release type to compare.
        b: The second release type to compare.
    
    Returns:
        True if release b is greater than release a.
        False otherwise.
    """
    pri1 = RELEASE_TYPE_PRIORITIES[a]
    pri2 = RELEASE_TYPE_PRIORITIES[b]
    return pri1 < pri2


class SemverRelease:
    """Class representing a release."""

    def __init__(self, config: SemverConfig):
        """Initialize SemverRelease object.
        
        Args:
            config: The SemverConfig to dictate behavior.
        """
        self._config = config

    def compute_release_type(self, commits: List[Commit]) -> ReleaseType:
        """Compute release type (major/minor/patch) based on commits.
        
        Args:
            commits: List of commits since last tag.
        
        Returns:
            ReleaseType that should be generated based on these commits.
        """
        release_type = ReleaseType.NO_RELEASE
        for commit in commits:
            for prefix in self._config.prefixes:
                if commit.message.startswith(f"{prefix.label}: "):
                    # Check whether this is greater than the existing value
                    if release_greater(release_type, prefix.type):
                        release_type = prefix.type
        return release_type

    def get_matching_tags(self, tags: List[str]) -> List[Tag]:
        """Parse git tags and return Tag objects.
        
        Only Git tags that match the `tag_format` in the config
        will be returned.

        Args:
            tags: Full list of Git tags for current repo.
        
        Returns:
            List of relevant tags, parsed into Tag objects.
        """
        pattern = (
            self._config.format.replace("%M", r"(?P<major>\d+)")
            .replace("%m", r"(?P<minor>\d+)")
            .replace("%p", r"(?P<patch>\d+)")
            .replace(".", r"\.")
        )
        matches = []
        for tag in tags:
            match = re.match(pattern, tag)
            if match:
                matches.append(
                    Tag(
                        name=tag,
                        major=int(match.group("major")),
                        minor=int(match.group("minor")),
                        patch=int(match.group("patch")),
                    )
                )
        return matches

    def get_biggest_tag(self, tags: List[str]) -> Optional[Tag]:
        """Among existing tags, determine which is newest.
        
        Args:
            tags: Full list of Git tags.
        
        Returns:
            Largest tag if at least one tag was provided.
            None if the `tags` arg was an empty list.
        """
        tags = self.get_matching_tags(tags)
        max_tag = None
        for tag in tags:
            if (
                max_tag is None
                or tag.major > max_tag.major
                or (tag.major == max_tag.major and tag.minor > max_tag.minor)
                or (
                    tag.major == max_tag.major
                    and tag.minor == max_tag.minor
                    and tag.patch > max_tag.patch
                )
            ):
                max_tag = tag
        return max_tag

    def compute_next_version(
        self, release_type: ReleaseType, tags: List[Tag]
    ) -> Tag:
        """Given the release type and tags, determine new tag (if any).
        
        Args:
            release_type: ReleaseType for next version.
            tags: List of Git tags in repo.
        
        Returns:
            Tag representing new version if new release needed.
            Biggest tag among existing tags if no need release needed.
        """
        if len(tags) < 1:
            name = (
                self._config.format.replace("%M", "0")
                .replace("%m", "1")
                .replace("%p", "0")
            )
            return Tag(name, 0, 1, 0)
        biggest_tag = self.get_biggest_tag(tags)
        if biggest_tag is None:
            max_version = (0, 1, 0)
        elif release_type == ReleaseType.MAJOR:
            max_version = (biggest_tag.major + 1, 0, 0)
        elif release_type == ReleaseType.MINOR:
            max_version = (biggest_tag.major, biggest_tag.minor + 1, 0)
        elif release_type == ReleaseType.PATCH:
            max_version = (biggest_tag.major, biggest_tag.minor, biggest_tag.patch + 1)
        else:
            max_version = (
                biggest_tag.major,
                biggest_tag.minor,
                biggest_tag.patch,
            )  # NO_RELEASE

        name = (
            self._config.format.replace("%M", str(max_version[0]))
            .replace("%m", str(max_version[1]))
            .replace("%p", str(max_version[2]))
        )
        return Tag(name, max_version[0], max_version[1], max_version[2])



"""Module for file replacer."""

import json
import os
import shutil
import toml
import yaml
from pagekey_semver.models import (
    JsonReplaceFile,
    ReplaceFileType,
    ReplaceFileUnion,
    SedReplaceFile,
    SemverConfig,
    TomlReplaceFile,
    YamlReplaceFile,
)
from pagekey_semver.release import Tag
from pagekey_semver.util.update_dict import set_dict_value


class FileReplacer:
    """Class to handle replacing files on new release."""

    def __init__(self, config: SemverConfig, new_version: Tag) -> None:
        """Initialize file replacer.
        
        Args:
            config: Settings for semver application.
            new_version: Latest version to replace file values with.
        """
        self._config = config
        self._new_version = new_version

    def replace_all(self) -> None:
        """Replace every file specified in config."""
        for replace_file in self._config.replace_files:
            self.replace_one(replace_file)

    def replace_one(self, file: ReplaceFileUnion) -> None:
        """Given a ReplaceFile, perform replace based on its type."""
        if file.type == ReplaceFileType.JSON:
            self.replace_json(file)
        elif file.type == ReplaceFileType.SED:
            self.replace_sed(file)
        elif file.type == ReplaceFileType.TOML:
            self.replace_toml(file)
        else:
            self.replace_yaml(file)

    def replace_json(self, replace_file: JsonReplaceFile) -> None:
        """Replace a JSON file's key with the new_version's name.
        
        Parses to JSON file into a dict, replaces the key,
        and dumps the result back into the original file.
        """
        with open(replace_file.name, "r") as replace_file_handle:
            contents = json.load(replace_file_handle)

        set_dict_value(contents, replace_file.key, self._new_version.name)

        with open(replace_file.name, "w") as replace_file_handle:
            json.dump(contents, replace_file_handle)

    def replace_sed(self, replace_file: SedReplaceFile):
        """Use sed to make an arbitrary replacement based on new version.
        
        Replaces %M/%m/%p with new_version's major/minor/patch,
        then calls sed with the provided `script`.
        """
        if shutil.which("sed") is None:
            raise EnvironmentError(
                "Sed executable not found on system - have you installed sed?"
            )
        script_escaped = replace_file.script.replace("'", "\\'")
        script_replaced = (
            script_escaped.replace("%M", str(self._new_version.major))
            .replace("%m", str(self._new_version.minor))
            .replace("%p", str(self._new_version.patch))
        )
        exit_code = os.system(f"sed -i '{script_replaced}' {replace_file.name}")
        if exit_code != 0:
            raise OSError("Invoking sed failed!")

    def replace_toml(self, replace_file: TomlReplaceFile):
        """Replace a TOML file's key with the new_version's name.
        
        Parses to TOML file into a dict, replaces the key,
        and dumps the result back into the original file.
        """
        with open(replace_file.name, "r") as replace_file_handle:
            contents = toml.load(replace_file_handle)

        set_dict_value(contents, replace_file.key, self._new_version.name)

        with open(replace_file.name, "w") as replace_file_handle:
            toml.dump(contents, replace_file_handle)

    def replace_yaml(self, replace_file: YamlReplaceFile):
        """Replace a YAML file's key with the new_version's name.
        
        Parses to YAML file into a dict, replaces the key,
        and dumps the result back into the original file.
        """
        with open(replace_file.name, "r") as replace_file_handle:
            contents = yaml.safe_load(replace_file_handle)

        set_dict_value(contents, replace_file.key, self._new_version.name)

        with open(replace_file.name, "w") as replace_file_handle:
            yaml.dump(contents, replace_file_handle)
